#include "stdafx.h"
#include <initguid.h>
#include <setupapi.h>
#include "testatm.h"
#include "patmif.h"
#include "ntddpack.h"
#include "sockatm.h"
#include "atm.h"
////////////#define INITGUID
//////////#undef  DEFINE_GUID
//////////#define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
//////////        EXTERN_C const GUID DECLSPEC_SELECTANY name \
//////////                = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }
#include "public.h"
#include <conio.h>
#include "common.h"

#define MAX_GUID_STRING_LEN   39          // 38 chars + terminator null


extern  CONNECTION_INFO ConnectionInfo;
extern  BOOL bStepMode;

static BOOL GetDeviceDescription( IN  LPTSTR DevPath,
						   OUT LPTSTR OutBuffer,
						   IN ULONG OutBufferLen )
{
    HDEVINFO                            hardwareDeviceInfo = NULL;
    SP_DEVICE_INTERFACE_DATA            deviceInterfaceData;
    SP_DEVINFO_DATA                     deviceInfoData;
    DWORD                               dwRegType, error;
    BOOL                                ret = FALSE;

    hardwareDeviceInfo = SetupDiCreateDeviceInfoList(NULL, NULL);
    if(INVALID_HANDLE_VALUE == hardwareDeviceInfo)
    {
        THROW_MESSAGE( "Couldn't create DeviceInfoList: %x", GetLastError());
        goto Error;
    }

    //
    // Enumerate devices of toaster class
    //
    deviceInterfaceData.cbSize = sizeof(deviceInterfaceData);

    SetupDiOpenDeviceInterface (hardwareDeviceInfo, DevPath,
                                 0, //
                                 &deviceInterfaceData);

    deviceInfoData.cbSize = sizeof(deviceInfoData);
    if(!SetupDiGetDeviceInterfaceDetail (
            hardwareDeviceInfo,
            &deviceInterfaceData,
            NULL, // probing so no output buffer yet
            0, // probing so output buffer length of zero
            NULL,
            &deviceInfoData) && (error = GetLastError()) != ERROR_INSUFFICIENT_BUFFER)
    {
        THROW_MESSAGE("Couldn't get interface detail: %x", GetLastError());
        goto Error;
    }
    //
    // Get the friendly name for this instance, if that fails
    // try to get the device description.
    //

    if(!SetupDiGetDeviceRegistryProperty(hardwareDeviceInfo, &deviceInfoData,
                                     SPDRP_FRIENDLYNAME,
                                     &dwRegType,
                                     (BYTE*) OutBuffer,
                                     OutBufferLen,
                                     NULL))
    {
        if(!SetupDiGetDeviceRegistryProperty(hardwareDeviceInfo, &deviceInfoData,
                                     SPDRP_DEVICEDESC,
                                     &dwRegType,
                                     (BYTE*) OutBuffer,
                                     OutBufferLen,
                                     NULL))
		{
            THROW_MESSAGE("Couldn't get friendlyname: %x", GetLastError());
            goto Error;
        }
    }

    ret = TRUE;

Error:
   if(hardwareDeviceInfo)
	{
        SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
    }

	return ret;
}



BOOL WDMGetATMDevices(PATM_ADAPTER pAdapters,DWORD *pAdapterCount)
{
    HDEVINFO                            hardwareDeviceInfo;
    SP_DEVICE_INTERFACE_DATA            deviceInterfaceData;
    PSP_DEVICE_INTERFACE_DETAIL_DATA    deviceInterfaceDetailData = NULL;
    ULONG                               predictedLength = 0;
    ULONG                               requiredLength = 0, i, j;

	GUID        InterfaceGuid;

	InterfaceGuid = GUID_DEVINTERFACE_PROATM;

    int DeviceIndex = 0;

    hardwareDeviceInfo = SetupDiGetClassDevs (
                       (LPGUID)&InterfaceGuid,
                       NULL, // Define no enumerator (global)
                       NULL, // Define no
                       (DIGCF_PRESENT | // Only Devices present
                       DIGCF_DEVICEINTERFACE)); // Function class devices.
    if(INVALID_HANDLE_VALUE == hardwareDeviceInfo)
    {
        goto Error;
    }

    //
    // Enumerate devices of a specific interface class
    //
    deviceInterfaceData.cbSize = sizeof(deviceInterfaceData);

	j = *pAdapterCount;
    for(i=0; SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
                                 0, // No care about specific PDOs
                                 (LPGUID)&InterfaceGuid,
                                 i, //
                                 &deviceInterfaceData) && j <MAX_ATM_ADAPTERS; i++ ) 
	{

        //
        // Allocate a function class device data structure to
        // receive the information about this particular device.
        //

        //
        // First find out required length of the buffer
        //
        if (deviceInterfaceDetailData) 
		{
            HeapFree (GetProcessHeap(), 0, deviceInterfaceDetailData);
        }

        if(!SetupDiGetDeviceInterfaceDetail (
                hardwareDeviceInfo,
                &deviceInterfaceData,
                NULL, // probing so no output buffer yet
                0, // probing so output buffer length of zero
                &requiredLength,
                NULL) && (GetLastError()) != ERROR_INSUFFICIENT_BUFFER)
        {
            goto Error;
        }
        predictedLength = requiredLength;

        deviceInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
                                        predictedLength);
        if (deviceInterfaceDetailData == NULL) 
		{
            goto Error;
        }

        deviceInterfaceDetailData->cbSize =  sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);


        if (! SetupDiGetDeviceInterfaceDetail (
                   hardwareDeviceInfo,
                   &deviceInterfaceData,
                   deviceInterfaceDetailData,
                   predictedLength,
                   &requiredLength,
                   NULL))
		{
            goto Error;
        }
        
	    if(!GetDeviceDescription( deviceInterfaceDetailData->DevicePath,
							      pAdapters[j].DriverDesc,
								  sizeof(pAdapters[j].DriverDesc)))
		{
			THROW_MESSAGE( "GetDeviceDescription failed %x", GetLastError());
			goto Error;
		}


		//
		// Copy the device path so that we can open the device using CreateFile.
		//
		strncpy( pAdapters[j].PnpInstanceID, deviceInterfaceDetailData->DevicePath, sizeof(pAdapters[j].PnpInstanceID));
		pAdapters[j].AdapterType = WDM_ADAPTER; 
		j++;
    }

	*pAdapterCount =j;
    if(deviceInterfaceDetailData)
	{
        HeapFree (GetProcessHeap(), 0, deviceInterfaceDetailData);
    }

    SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
    return TRUE;

Error:

	THROW_MESSAGE( "GetWDMATMDevices Failed. Error Code:%d", GetLastError());

    if(deviceInterfaceDetailData)
        HeapFree (GetProcessHeap(), 0, deviceInterfaceDetailData);

    SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
    return FALSE;
}

static BOOL WDMATMOpenAdapter( PATM_ADAPTER pAdapter)
{

    HANDLE                  hDevice;

    //
    // Open an handle to the device.
    //
    hDevice = CreateFile (
			pAdapter->PnpInstanceID,
            GENERIC_READ | GENERIC_WRITE,
            0,
            NULL, // no SECURITY_ATTRIBUTES structure
            OPEN_EXISTING, // No special create flags
            FILE_FLAG_OVERLAPPED,
            NULL);

    if (INVALID_HANDLE_VALUE == hDevice) 
	{
        THROW_MESSAGE("Failed to open the device: %ws", pAdapter->PnpInstanceID);
        return FALSE;
    }

	pAdapter->hFile = hDevice;
	pAdapter->bCreatedDevice = TRUE;

    return TRUE;


}

static BOOL WDMATMCloseAdapter( PATM_ADAPTER pAdapter)
{
 	if(pAdapter->hFile)
	{
		CloseHandle( pAdapter->hFile );
		pAdapter->hFile=NULL;
	}

	pAdapter->bCreatedDevice = FALSE;

	return TRUE;
}

BOOL WDMATMGetInformation ( PATM_ADAPTER pAdapter)
{
	LPPROTOCOL_DRIVER lpProtocolATM=NULL;
    ULONG RequestLength;
    BOOL bResult = FALSE;
    PATM_APPLICATION_INFO pInfo;

	if (!WDMATMOpenAdapter(pAdapter))
		return FALSE;

	if (pAdapter->bCreatedDevice)
    {
	    lpProtocolATM = ProtocolATMAllocateData();
        RequestLength = offsetof(PROTOCOL_OID_DATA, Data) + sizeof(ATM_APPLICATION_INFO);

	    if ( lpProtocolATM && ProtocolATMInitData( lpProtocolATM,RequestLength) )
	    {

	        #define pOidData ((PPROTOCOL_OID_DATA)(lpProtocolATM->Buffer))

		    pOidData->Oid    = OID_ATM_APPLICATION_INFO;
		    pOidData->Length = sizeof(ATM_APPLICATION_INFO );

	        bResult = ProtocolATMDeviceIoControl( pAdapter,
									             lpProtocolATM,
									             IOCTL_ATM_QUERY_OID_VALUE,
									             TRUE );

	        if ( bResult && (lpProtocolATM->ulBytesReceived >= RequestLength) )
	        {
		        pInfo  = (PATM_APPLICATION_INFO)pOidData->Data;
                pAdapter->MemorySize = pInfo->MemorySize;
                pAdapter->Speed      = pInfo->Speed;
                pAdapter->LinkStatus = pInfo->LinkStatus;
                memcpy(&pAdapter->MacAddress, pInfo->MacAddress,6);
                pAdapter->DriverVersion = pInfo->DriverVersion;
         	    bResult = TRUE;
   
            }
            else
            {
                THROW_MESSAGE("Error in function <DeviceIoControl>",0);
            }
     
	    }
        else
        {
            THROW_MESSAGE("Error in function <ProtocolATMAllocateData>",0);

        }
        ProtocolATMFreeData(lpProtocolATM);
    }

	WDMATMCloseAdapter(pAdapter);

    return bResult;

}


BOOL	 WDMATMSetLoopback    ( PATM_ADAPTER pAdapter, BOOL bLoopback)
{
	LPPROTOCOL_DRIVER lpProtocolATM=NULL;
    ULONG RequestLength;
    BOOL bResult = FALSE;

	if (!WDMATMOpenAdapter(pAdapter))
		return FALSE;

    if (pAdapter->bCreatedDevice)
    {
	    lpProtocolATM = ProtocolATMAllocateData();
        RequestLength = offsetof(PROTOCOL_OID_DATA, Data) + 4;

	    if ( lpProtocolATM && ProtocolATMInitData( lpProtocolATM,RequestLength) )
	    {

	        #define pOidData ((PPROTOCOL_OID_DATA)(lpProtocolATM->Buffer))


			pOidData->Oid    = OID_ATM_SET_LOOPBACK_MODE;

			*(ULONG *)(pOidData->Data) = bLoopback;
		    pOidData->Length = 4;

	        bResult = ProtocolATMDeviceIoControl( pAdapter,
									             lpProtocolATM,
									             IOCTL_ATM_SET_OID_VALUE,
									             TRUE );

	        if ( !bResult)
	        {
                THROW_MESSAGE("Error in function <DeviceIoControl>",0);
            }
        
      
	    }
        else
        {
            THROW_MESSAGE("Error in function <ProtocolATMAllocateData>",0);

        }
        ProtocolATMFreeData(lpProtocolATM);
    }

	WDMATMCloseAdapter(pAdapter);

    return bResult;

}

// Initializes a PVC Connection
static int WDMCreatePVC( PATM_ADAPTER pAdapter,
						 PATM_SETTINGS pSettings)
{

	LPPROTOCOL_DRIVER   lpProtocolATM=NULL;
	ATM_PARAMETERS	AtmParameters;
	BOOL bResult= FALSE;
	UINT  RequestLength;

	if (!WDMATMOpenAdapter(pAdapter))
		return SOCKET_ERROR;

     if (pAdapter->bCreatedDevice)
    {
		memset( &AtmParameters, 0, sizeof(ATM_PARAMETERS)) ;

		if (pSettings->ReceiveFlow)
			AtmParameters.Flags	  |=  ENABLE_RECEIVE;
		if (pSettings->XmitFlow)
			AtmParameters.Flags	  |=  ENABLE_TRANSMIT;
		
		if (LAYER_AAL5 == pSettings->AalType )
			AtmParameters.AALType = AAL_TYPE_AAL5;
		else
			AtmParameters.AALType = AAL_TYPE_AAL0;

		AtmParameters.ConnectionId.Vpi = pSettings->VcSettings.Pvc.Vpi;
		AtmParameters.ConnectionId.Vci = pSettings->VcSettings.Pvc.Vci;

		if ( pSettings->QualityOfServices == QOS_UBR)
		{
			AtmParameters.Transmit.ServiceCategory = ATM_SERVICE_CATEGORY_UBR;
			AtmParameters.Transmit.PeakCellRate    = pSettings->QosSettings.Ubr.PeakCellRate;
			AtmParameters.Transmit.MaxSduSize	   = pSettings->PacketSize;
		}
		else if ( pSettings->QualityOfServices == QOS_CBR)
		{
			AtmParameters.Transmit.ServiceCategory = ATM_SERVICE_CATEGORY_CBR;
			AtmParameters.Transmit.PeakCellRate    = pSettings->QosSettings.Cbr.PeakCellRate;
			AtmParameters.Transmit.MaxSduSize	   = pSettings->PacketSize;

		}
		else if ( pSettings->QualityOfServices == QOS_VBR)
		{

			AtmParameters.Transmit.ServiceCategory  = ATM_SERVICE_CATEGORY_CBR;
			AtmParameters.Transmit.PeakCellRate     = pSettings->QosSettings.Vbr.PeakCellRate;
			AtmParameters.Transmit.AverageCellRate  = pSettings->QosSettings.Vbr.SubstainableCellRate;
			AtmParameters.Transmit.BurstLengthCells = pSettings->QosSettings.Vbr.MaxBurstSize;

		}

		RequestLength =	 sizeof(ATM_PARAMETERS);

		lpProtocolATM = ProtocolATMAllocateData();
		if ( lpProtocolATM  && ProtocolATMInitData( lpProtocolATM, RequestLength) )
		{

			memcpy( lpProtocolATM->Buffer, &AtmParameters, sizeof(ATM_PARAMETERS));
			bResult = ProtocolATMDeviceIoControl( pAdapter,
												  lpProtocolATM,
												  IOCTL_ATM_CREATE_VC,
												  TRUE );
	    }
        ProtocolATMFreeData(lpProtocolATM);
    }

	if (!bResult)
	{
   		THROW_MESSAGE("Error: WDMCreatePVC. ErrorCode:%d", GetLastError());
		WDMATMCloseAdapter(pAdapter);
		return SOCKET_ERROR;
	}

	ConnectionInfo.Vpi          = pSettings->VcSettings.Pvc.Vpi;
	ConnectionInfo.Vci          = pSettings->VcSettings.Pvc.Vci;
	ConnectionInfo.DeviceNumber = pSettings->DeviceNumber;
	ConnectionInfo.AALLayer     = pSettings->AalType;
	ConnectionInfo.Gfc          = pSettings->VcSettings.Pvc.Gfc;
	ConnectionInfo.Pt           = pSettings->VcSettings.Pvc.Pt;
	ConnectionInfo.Clp          = pSettings->VcSettings.Pvc.Clp;
	ConnectionInfo.PacketSize   = pSettings->PacketSize;	

	return (int)pAdapter->hFile;
}




// The Receive Thread
static DWORD RecvThread(DWORD param)
{
	DWORD  bytesRecv;
	PPACKET_HEADER pPacketHeader;
    int i,iError;
	PAAL0_CELL pCell;
	DWORD dwEvent;

	int CurrentIndex=0;
	

	struct timeval timeselect={1,0};

    ConnectionInfo.RecvSeq =0;
	iError = 0;	


	for (i=0; i < MAX_RECV_PACKETS; i++)
	{
		ConnectionInfo.RecvPacket[i].len = ConnectionInfo.PacketSize;
		if (!ReadFile( ConnectionInfo.Socket,	ConnectionInfo.RecvPacket[i].buf, ConnectionInfo.RecvPacket[i].len,
			             &bytesRecv, &ConnectionInfo.RecvOverlapped[i] ))
		{
			iError =GetLastError();
			if (iError != ERROR_IO_PENDING)
			{
               THROW_MESSAGE("Error: ReadFile. ErrorCode:%d", iError);
 				break;
			}

			ConnectionInfo.hRecvEventList[i]=ConnectionInfo.RecvOverlapped[i].hEvent;
		}
	}
 
	if (iError != 0 && iError != ERROR_IO_PENDING)
	{
		THROW_MESSAGE("Error: RecvThread. ErrorCode:%d", iError);
		if (ConnectionInfo.DialogHandle)
			PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
	}
	else
	{
		ConnectionInfo.hRecvEventList[MAX_RECV_PACKETS]=ConnectionInfo.RecvCloseEvent;
		while(TRUE)
		{
			dwEvent = WaitForMultipleObjects(1 + MAX_RECV_PACKETS - CurrentIndex, &ConnectionInfo.hRecvEventList[CurrentIndex] , FALSE, INFINITE);


			if (dwEvent <  (WAIT_OBJECT_0 + MAX_RECV_PACKETS - CurrentIndex))
			{ 
				CurrentIndex +=dwEvent;
//				ResetEvent(ConnectionInfo.RecvOverlapped[CurrentIndex].hEvent);
				while (GetOverlappedResult(ConnectionInfo.Socket,&ConnectionInfo.RecvOverlapped[CurrentIndex], &bytesRecv,FALSE) == TRUE)
				{
					if (ConnectionInfo.AALLayer != LAYER_AAL0)
						pPacketHeader = (PPACKET_HEADER) ConnectionInfo.RecvPacket[CurrentIndex].buf;
					else
					{
						pCell = (PAAL0_CELL)ConnectionInfo.RecvPacket[CurrentIndex].buf;
						ConnectionInfo.RecvCellHeader = ntohl(pCell->CellHeader);
						pPacketHeader = (PPACKET_HEADER)pCell->Payload;
					}
						
					if (pPacketHeader->seq != ConnectionInfo.RecvSeq)
					{
						if (ConnectionInfo.RecvSeq)
							ConnectionInfo.pktDrop += (pPacketHeader->seq - ConnectionInfo.RecvSeq);
					}
					ConnectionInfo.RecvSeq = pPacketHeader->seq + 1;
					ConnectionInfo.pktRecv ++;
					ConnectionInfo.bytesRecv += bytesRecv;

					if (!ReadFile( ConnectionInfo.Socket, ConnectionInfo.RecvPacket[CurrentIndex].buf, ConnectionInfo.RecvPacket[CurrentIndex].len,
			             &bytesRecv, &ConnectionInfo.RecvOverlapped[CurrentIndex] ))

					{
						iError =GetLastError();
						if (iError != ERROR_IO_PENDING)
							break;
					}
					CurrentIndex = (CurrentIndex +1) %MAX_RECV_PACKETS;

				}
				if (0)
//				if ( ((iError = GetLastError()) !=ERROR_IO_INCOMPLETE) && (iError != ERROR_IO_PENDING))
				{
					THROW_MESSAGE("Error: GetOverlappedResult. ErrorCode:%d", iError);
					if (ConnectionInfo.DialogHandle)
       					PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
					break;
				}
	        
			}
			else if (dwEvent ==  WAIT_OBJECT_0 + MAX_RECV_PACKETS - CurrentIndex )
			{
				// close event
				break;
			}

			else
			{
				THROW_MESSAGE("Error: WaitForMultipleEvents. ErrorCode:%d", GetLastError());
				if (ConnectionInfo.DialogHandle)
       				PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
				break;
	        
			}
		}
	}
	ConnectionInfo.RecvThreadHandle = NULL;
    return SUCCESS;
}



// The Transmit Thread
static DWORD XmitThread(DWORD param)
{
	DWORD  bytesXmit, XmitFlags;
	PPACKET_HEADER pPacketHeader;
	PAAL0_CELL pCell;
    int i, iError;
	DWORD dwEvent, PacketIndex;

	int  XmitBurst = MAX_SEND_PACKETS;
	if (bStepMode)
		  XmitBurst = 1;

	struct timeval timeselect={1,0};

	for (i=0; i < XmitBurst; i++)
	{
		if (ConnectionInfo.AALLayer != LAYER_AAL0)
		{
			pPacketHeader = (PPACKET_HEADER) ConnectionInfo.XmitPacket[i].buf;
			pPacketHeader->seq = i;
			pPacketHeader->len = ConnectionInfo.XmitPacket[i].len;
	
		}
		else

		{
			pCell = (PAAL0_CELL)ConnectionInfo.XmitPacket[i].buf;

			pCell->CellHeader =     (((ConnectionInfo.Gfc << 28) & 0xF0000000) |  
			                    ((ConnectionInfo.Vpi << 20) & 0x0FF00000) | 
						        ((ConnectionInfo.Vci << 4 ) & 0x000FFFF0) | 
						        ((ConnectionInfo.Pt  << 1)  & 0x0000000E) | 
						        (ConnectionInfo.Clp         & 0x00000001));
			pCell->CellHeader = htonl(pCell->CellHeader);

//$$$$$$$$$$ for OAM test puposes $$$$$$$$$$$$$
//		pCell->Payload[0]=0x81;
//	    pCell->Payload[1]=0x5;
//		memset(&pCell->Payload[2], 0,46);
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

			pPacketHeader = (PPACKET_HEADER)pCell->Payload;
			pPacketHeader->len = 52;
			pPacketHeader->seq =i;
		}

		if(!WriteFile( ConnectionInfo.Socket, ConnectionInfo.XmitPacket[i].buf, ConnectionInfo.XmitPacket[i].len,
			          &bytesXmit,&ConnectionInfo.XmitOverlapped[i]))
		{

            iError = GetLastError();
            if (iError != ERROR_IO_PENDING)
                break;

			ConnectionInfo.hXmitEventList[i]=ConnectionInfo.XmitOverlapped[i].hEvent;
		}
	}

  	if (iError != 0 && iError != ERROR_IO_PENDING)
	{
		THROW_MESSAGE("Error: WriteFile. ErrorCode:%d", iError);
		if (ConnectionInfo.DialogHandle)
			PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
	}
	else
	{

  
		ConnectionInfo.hXmitEventList[XmitBurst]=ConnectionInfo.XmitCloseEvent;
	    PacketIndex = XmitBurst;
		while(TRUE)
		{	
		
			XmitFlags = ZERO_FLAGS;
			bytesXmit=ConnectionInfo.XmitPacket[0].len;		  // Assume all packets have the same size

	        
			dwEvent = WaitForMultipleObjects(XmitBurst+1,ConnectionInfo.hXmitEventList , FALSE, INFINITE);

			if (dwEvent ==  WAIT_OBJECT_0 + XmitBurst)
			{
				// close event 
				break;
			}

			else if (dwEvent <  (WAIT_OBJECT_0 + XmitBurst ))
			{ 

				while (GetOverlappedResult(ConnectionInfo.Socket,&ConnectionInfo.XmitOverlapped[dwEvent],&bytesXmit,FALSE) == TRUE)
				{
					if (ConnectionInfo.AALLayer != LAYER_AAL0)
						pPacketHeader = (PPACKET_HEADER) ConnectionInfo.XmitPacket[dwEvent].buf;
					else
					{
						pCell = (PAAL0_CELL)ConnectionInfo.XmitPacket[dwEvent].buf;
						pPacketHeader = (PPACKET_HEADER)pCell->Payload;
					}

					ConnectionInfo.pktXmit = pPacketHeader->seq;
					ConnectionInfo.bytesXmit += bytesXmit;
					pPacketHeader->seq = PacketIndex++;			

					if (bStepMode)
						if (AfxMessageBox(" One packet send . Press YES to resend or NO to abort", MB_YESNO |MB_USERICON ) == IDNO)
							break;

					if(!WriteFile( ConnectionInfo.Socket, ConnectionInfo.XmitPacket[dwEvent].buf, ConnectionInfo.XmitPacket[dwEvent].len,
			          &bytesXmit,&ConnectionInfo.XmitOverlapped[dwEvent]))
					{

						iError = GetLastError();
							if (iError != ERROR_IO_PENDING)
								break;
					}
					dwEvent =   (dwEvent + 1)% XmitBurst;
		  			if ( WaitForSingleObject(ConnectionInfo.XmitCloseEvent, 0)== WAIT_OBJECT_0 )
						goto End;						
 

				}
				iError = GetLastError();
				if (iError != ERROR_IO_INCOMPLETE)
				{
					THROW_MESSAGE("Error: GetOverlappedResult. ErrorCode:%d", iError);
					if (ConnectionInfo.DialogHandle)
  						PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
						break;
				}
			}
			else
			{
				THROW_MESSAGE("Error: WaitForMultipleObjectss. ErrorCode:%d", GetLastError());
				if (ConnectionInfo.DialogHandle)
  					PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
				break;
			}
		}
    }
End:	
    ConnectionInfo.XmitThreadHandle = NULL;
	return SUCCESS;
}


 // Stops the Receive Thread
static void StopReceiveMode()
{
	int i;
    if (ConnectionInfo.RecvThreadHandle && ConnectionInfo.RecvCloseEvent)
    {
        SetEvent(ConnectionInfo.RecvCloseEvent);
        
         // Wait 5 seconds for it to exit.  Shouldn't happen.
        if (WaitForSingleObject(ConnectionInfo.RecvThreadHandle, 5000) == WAIT_TIMEOUT)
        {
            TerminateThread(ConnectionInfo.RecvThreadHandle, 0);
        }

    }


	for (i=0; i <MAX_RECV_PACKETS; i++)
	{

		if (ConnectionInfo.RecvOverlapped[i].hEvent)
		{
			CloseHandle(ConnectionInfo.RecvOverlapped[i].hEvent);
			ConnectionInfo.RecvOverlapped[i].hEvent = 0;
		}

		if (ConnectionInfo.RecvPacket[i].buf)
		{
            free(ConnectionInfo.RecvPacket[i].buf);    
            ConnectionInfo.RecvPacket[i].buf = NULL;
		}
	}

   
    if (ConnectionInfo.RecvCloseEvent)
    {
        CloseHandle(ConnectionInfo.RecvCloseEvent);
        ConnectionInfo.RecvCloseEvent = 0;
    }

}


// Starts the  Receive Thread
static BOOL StartReceiveMode(void)
{
		int i;
    ConnectionInfo.RecvMode   = TRUE;    

    
    ConnectionInfo.RecvCloseEvent = CreateEvent(0, FALSE, FALSE, NULL);
    if (ConnectionInfo.RecvCloseEvent == NULL)
    {
   		THROW_MESSAGE("Error: CreateEvent. ErrorCode:%d", GetLastError());
        return FALSE;
    }
 
	for (i = 0; i <MAX_RECV_PACKETS; i++)
	{
		ConnectionInfo.RecvOverlapped[i].hEvent= CreateEvent(0, FALSE, FALSE, NULL);
		if (ConnectionInfo.RecvOverlapped[i].hEvent == NULL)
		{   
   		    THROW_MESSAGE("Error: CreateEvent. ErrorCode:%d", GetLastError());
            return FALSE;
		}

		if ((ConnectionInfo.RecvPacket[i].buf = (char *)malloc(ConnectionInfo.PacketSize) ) == NULL)
		{
   			THROW_MESSAGE("Error: malloc", 0);
			return FALSE;
		}

	}



  	if ((ConnectionInfo.RecvThreadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)RecvThread, 0, 0, &ConnectionInfo.RecvThreadID)) == 0)
    {
   
   		THROW_MESSAGE("Error: CreateThread(RecvThread ..). ErrorCode:%d", GetLastError());
        return FALSE;
    }
	SetThreadPriority( ConnectionInfo.RecvThreadHandle, THREAD_PRIORITY_HIGHEST );
    return TRUE;

}


// Stops the Transmit Thread
static void StopXmitMode()
{
	int i;

    if (ConnectionInfo.XmitThreadHandle && ConnectionInfo.XmitCloseEvent)
    {
        SetEvent(ConnectionInfo.XmitCloseEvent);
        
         // Wait 5 seconds for it to exit.  Shouldn't happen.
        if (WaitForSingleObject(ConnectionInfo.XmitThreadHandle, 5000) == WAIT_TIMEOUT)
        {
            TerminateThread(ConnectionInfo.XmitThreadHandle, 0);
        }

    }

	for (i=0; i<MAX_SEND_PACKETS; i++)
	{
		if (ConnectionInfo.XmitOverlapped[i].hEvent)
		{
			CloseHandle(ConnectionInfo.XmitOverlapped[i].hEvent);
			ConnectionInfo.XmitOverlapped[i].hEvent = 0;
		}

		if (ConnectionInfo.XmitPacket[i].buf)
		{
            free(ConnectionInfo.XmitPacket[i].buf);    
            ConnectionInfo.XmitPacket[i].buf = NULL;
		}
	}
      
    if (ConnectionInfo.XmitCloseEvent)
    {
        CloseHandle(ConnectionInfo.XmitCloseEvent);
        ConnectionInfo.XmitCloseEvent = 0;
    }
 
}


static BOOL StartXmitMode(void)
{
    int i;
    
    ConnectionInfo.XmitMode   = TRUE;    
 
	for (i=0; i <MAX_SEND_PACKETS; i++)
	{
		if ((ConnectionInfo.XmitPacket[i].buf = (char *)malloc(ConnectionInfo.PacketSize) ) == NULL)
		{
			THROW_MESSAGE("Error: malloc", 0);
			return FALSE;
		}
		ConnectionInfo.XmitPacket[i].len = ConnectionInfo.PacketSize; 

		ConnectionInfo.XmitOverlapped[i].hEvent=CreateEvent(0, FALSE, FALSE, NULL);
	   
		if (ConnectionInfo.XmitOverlapped[i].hEvent == NULL)
		{   
   			THROW_MESSAGE("Error: CreateEvent. ErrorCode:%d", GetLastError());
			return FALSE;
		}

	}

    ConnectionInfo.XmitCloseEvent = CreateEvent(0, FALSE, FALSE, NULL);
    if (ConnectionInfo.XmitCloseEvent == NULL)
    {
        
   		THROW_MESSAGE("Error: CreateEvent. ErrorCode:%d", GetLastError());
        return FALSE;
    }
 
    


  	if ((ConnectionInfo.XmitThreadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)XmitThread, 0, 0, &ConnectionInfo.XmitThreadID)) == 0)
    {
   		THROW_MESSAGE("Error: CreateThread(XmitThread ..). ErrorCode:%d", GetLastError());
        return FALSE;
    }

    return TRUE;
}

void WDMStopATMConnection(void)
{
    if(ConnectionInfo.XmitMode)
        StopXmitMode();

    if (ConnectionInfo.RecvMode)
        StopReceiveMode();
            
	WDMATMCloseAdapter( ConnectionInfo.pAdapter);

    memset(&ConnectionInfo,0,sizeof(CONNECTION_INFO));

}


int WDMStartATMConnection(PATM_ADAPTER pAdapter, PATM_SETTINGS pSettings,HWND DialogHandle)
{
    memset(&ConnectionInfo,0,sizeof(CONNECTION_INFO));

    ConnectionInfo.DialogHandle = DialogHandle;
	ConnectionInfo.pAdapter     = pAdapter;
    if (pSettings->VcType == CONNECTION_PVC)
    {

                
        if ((ConnectionInfo.Socket=(HANDLE)WDMCreatePVC(pAdapter, pSettings) ) == (HANDLE)SOCKET_ERROR)
        {
            return STATE_NOT_CONNECTED;
            
        }
        ConnectionInfo.ConnectionState = STATE_CONNECTED;

    }
    ConnectionInfo.VcType = pSettings->VcType;
    

	if ( pSettings->ReceiveFlow)
    {
        if (StartReceiveMode() == FALSE)
        {
            WDMStopATMConnection();
            return STATE_NOT_CONNECTED;
        }
    }

    if ( pSettings->XmitFlow)
    {
        if (StartXmitMode() == FALSE)
        {
            SockStopATMConnection();
            return STATE_NOT_CONNECTED;
        }
    }

	return ConnectionInfo.ConnectionState;

}

